import functools
import re
from datetime import datetime

import MetaTrader5 as _mt5
import numpy

from . import const as _const
from . import helpers as _h
from .state import global_state as _state
from .types import *


class MT5Error(Exception):
    """The exception class for all MetaTrader5 Errors.

    Example:
        >>> try:
        >>>     raise MT5Error(_const.ERROR_CODE.UNSUPPORTED, "This feature is unsupported.")
        >>> except MT5Error as e:
        >>>     print(e.error_code, e.description)


    """
    def __init__(self, error_code: _const.ERROR_CODE, description: str):
        """

        :param error_code: int error code returned my last_error()
        :param description: error description
        """
        super().__init__(f"{error_code.name}: {description}")
        self.errno = self.error_code = error_code
        self.strerror = self.description = description


def _context_manager_modified(f):
    @functools.wraps(f)
    def cmm_wrapped_func(*args, **kwargs):
        result = f(*args, **kwargs)
        last_err = None
        if _state.debug_logging:
            last_err = mt5_last_error()
            call_sig = f"{f.__name__}({_h.args_to_str(args, kwargs)})"
            _state.logger(f"[{call_sig}][{last_err}]")
        # make sure we logger before we raise
        if _state.raise_on_errors:  # no need to check last error if we got a result
            if isinstance(result, numpy.ndarray):
                is_result = True if len(result) > 0 else False
            else:
                is_result = bool(result)
            if not is_result:
                error_code, description = last_err or mt5_last_error()
                if error_code != _const.ERROR_CODE.OK:
                    if error_code == _const.ERROR_CODE.INVALID_PARAMS:
                        description += str(args) + str(kwargs)
                    raise MT5Error(_const.ERROR_CODE(error_code), description)
        if _state.return_as_dict:
            result = _h.as_dict_all(result)
        return result

    return cmm_wrapped_func


mt5_account_info = _mt5.account_info


def parse_args(default_symbol:str=None,
               default_timeframe:Union[int, _const.TIMEFRAME]=None,
               ) -> Optional[Tuple[str, _const.TIMEFRAME]]:
    import sys
    try:
        symbol = sys.argv[1]
        arg2 = sys.argv[2]
        timeframe = _const.MINUTES_TO_TIMEFRAME.get(int(arg2))
        if timeframe is None:
            timeframe = _const.TIMEFRAME(arg2)
        return symbol, timeframe
    except Exception:
        if default_symbol and default_timeframe:
            return default_symbol, _const.TIMEFRAME(default_timeframe)
        if _state.raise_on_errors:
            raise MT5Error(_const.ERROR_CODE.INVALID_COMMANDLINE_ARGS, f"Invalid command line args {sys.argv}")
        else:
            return None


@_context_manager_modified
def account_info() -> AccountInfo:
    """Get info on the current trading account. The function returns all data that can be obtained using
    AccountInfoInteger, AccountInfoDouble and AccountInfoString in one call.

    :return: Return info in the form of a named tuple structure (namedtuple). Return None in case of an error.
    The info on the error can be obtained using last_error().
    """
    return mt5_account_info()


@_context_manager_modified
def copy_rates(symbol,
               timeframe: int,
               *,
               datetime_from: Union[datetime, int] = None,
               datetime_to: Union[datetime, int] = None,
               start_pos: int = None,
               count: int = None,
               ) -> Union[numpy.ndarray, None]:
    """Generic function to use keywords to automatically call the correct copy rates function depending on the
    keyword args passed in.

    :param symbol: Financial instrument name, for example, "EURUSD".
    :param timeframe: Timeframe the bars are requested for. Set by a value from the TIMEFRAME enumeration.
    :param datetime_from: Date of opening of the first bar from the requested sample. Set by the 'datetime'
    object or as a number of seconds elapsed since 1970.01.01.
    :param datetime_to: Date, up to which the bars are requested. Set by the 'datetime' object or as a number
    of seconds elapsed since 1970.01.01. Bars with the open time <= date_to are returned.
    :param start_pos: Initial index of the bar the data are requested from. The numbering of bars goes from
    present to past. Thus, the zero bar means the current one.
    :param count: Number of bars to receive.
    :return: Returns bars as the numpy array with the named time, open, high, low, close, tick_volume,
    spread and real_volume columns. Return None in case of an error. The info on the error can be obtained
    using last_error().
    """
    # TODO: test this without args and with count only
    symbol = _h.any_symbol(symbol)
    try:
        if datetime_from is not None:
            if count is not None:
                return mt5_copy_rates_from(symbol, timeframe, datetime_from, count)
            if datetime_to is not None:
                return mt5_copy_rates_range(symbol, timeframe, datetime_from, datetime_to)
        if all(x is None for x in [datetime_from, datetime_to, start_pos]):
            start_pos = 0
        count = min((count or _state.max_bars), _state.max_bars - 1)
        return mt5_copy_rates_from_pos(symbol, timeframe, start_pos, count)
    except SystemError:
        return None


mt5_copy_rates_from = _mt5.copy_rates_from


@_context_manager_modified
def copy_rates_from(symbol,
                    timeframe: int,
                    datetime_from: Union[datetime, int],
                    count: int
                    ) -> Union[numpy.ndarray, None]:
    """Get bars from the MetaTrader 5 terminal starting from the specified date.

    :param symbol: Financial instrument name, for example, "EURUSD".
    :param timeframe: Timeframe the bars are requested for. Set by a value from the TIMEFRAME enumeration.
    :param datetime_from: Date of opening of the first bar from the requested sample. Set by the 'datetime'
    object or as a number of seconds elapsed since 1970.01.01.
    :param count: Number of bars to receive.
    :return: Returns bars as the numpy array with the named time, open, high, low, close, tick_volume,
    spread and real_volume columns. Return None in case of an error. The info on the error can be obtained
    using last_error().
    """
    symbol = _h.any_symbol(symbol)
    try:
        return mt5_copy_rates_from(symbol, timeframe, datetime_from, count)
    except SystemError:
        return None


mt5_copy_rates_from_pos = _mt5.copy_rates_from_pos


@_context_manager_modified
def copy_rates_from_pos(symbol,
                        timeframe: int,
                        start_pos: int,
                        count: int
                        ) -> Union[numpy.ndarray, None]:
    """Get bars from the MetaTrader 5 terminal starting from the specified index.

    :param symbol: Financial instrument name, for example, "EURUSD".
    :param timeframe: Timeframe the bars are requested for. Set by a value from the TIMEFRAME enumeration.
    :param start_pos: Initial index of the bar the data are requested from. The numbering of bars goes from
    present to past. Thus, the zero bar means the current one.
    :param count: Number of bars to receive.
    :return: Returns bars as the numpy array with the named time, open, high, low, close, tick_volume,
    spread and real_volume columns. Return None in case of an error. The info on the error can be obtained
    using last_error().
    """
    symbol = _h.any_symbol(symbol)
    try:
        return mt5_copy_rates_from_pos(symbol, timeframe, start_pos, count)
    except SystemError:
        return None


mt5_copy_rates_range = _mt5.copy_rates_range


@_context_manager_modified
def copy_rates_range(symbol,
                     timeframe: int,
                     datetime_from: Union[datetime, int],
                     datetime_to: Union[datetime, int]
                     ) -> Union[numpy.ndarray, None]:
    """Get bars from the MetaTrader 5 terminal starting from the specified index.

        :param symbol: Financial instrument name, for example, "EURUSD".
        :param timeframe: Timeframe the bars are requested for. Set by a value from the TIMEFRAME enumeration.
        :param datetime_from: Date of opening of the first bar from the requested sample. Set by the 'datetime'
        object or as a number of seconds elapsed since 1970.01.01.
        :param datetime_to: Date, up to which the bars are requested. Set by the 'datetime' object or as a number
        of seconds elapsed since 1970.01.01. Bars with the open time <= date_to are returned.
        :return: Returns bars as the numpy array with the named time, open, high, low, close, tick_volume,
        spread and real_volume columns. Return None in case of an error. The info on the error can be obtained
        using last_error().
        """
    symbol = _h.any_symbol(symbol)
    try:
        return mt5_copy_rates_range(symbol, timeframe, datetime_from, datetime_to)
    except SystemError:
        return None


mt5_copy_ticks_from = _mt5.copy_ticks_from


@_context_manager_modified
def copy_ticks_from(symbol,
                    datetime_from: Union[datetime, int],
                    count: int,
                    flags: int,
                    ) -> Union[numpy.ndarray, None]:
    """Get ticks from the MetaTrader 5 terminal starting from the specified date.

    :param symbol: Financial instrument name, for example, "EURUSD". Required unnamed parameter.
    :param datetime_from: Date the ticks are requested from. Set by the 'datetime' object or as a number of
    seconds elapsed since 1970.01.01.
    :param count: Number of ticks to receive.
    :param flags: A flag to define the type of the requested ticks. COPY_TICKS_INFO – ticks with Bid and/or
    Ask changes, COPY_TICKS_TRADE – ticks with changes in Last and Volume, COPY_TICKS_ALL – all ticks.
    Flag values are described in the COPY_TICKS enumeration.
    :return: Returns ticks as the numpy array with the named time, bid, ask, last and flags columns.
    The 'flags' value can be a combination of flags from the TICK_FLAG enumeration. Return None in case of an error.
    The info on the error can be obtained using last_error().

    Note:
        See the CopyTicks function for more information.

        When creating the 'datetime' object, Python uses the local time zone, while MetaTrader 5 stores tick and
        bar open time in UTC time zone (without the shift). Therefore, 'datetime' should be created in UTC time
        for executing functions that use time. Data received from the MetaTrader 5 terminal has UTC time.
    """
    symbol = _h.any_symbol(symbol)
    try:
        return mt5_copy_ticks_from(symbol, datetime_from, count, flags)
    except SystemError:
        return None


mt5_copy_ticks_range = _mt5.copy_ticks_range


@_context_manager_modified
def copy_ticks_range(symbol,
                     datetime_from: Union[datetime, int],
                     datetime_to: Union[datetime, int],
                     flags: int,
                     ) -> Union[numpy.ndarray, None]:
    """Get ticks from the MetaTrader 5 terminal starting from the specified date.

        :param symbol: Financial instrument name, for example, "EURUSD". Required unnamed parameter.
        :param datetime_from: Date the ticks are requested from. Set by the 'datetime' object or as a number of
        seconds elapsed since 1970.01.01.
        :param datetime_to: Date, up to which the ticks are requested. Set by the 'datetime' object or as a number
        of seconds elapsed since 1970.01.01.
        :param flags: A flag to define the type of the requested ticks. COPY_TICKS_INFO – ticks with Bid and/or
        Ask changes, COPY_TICKS_TRADE – ticks with changes in Last and Volume, COPY_TICKS_ALL – all ticks.
        Flag values are described in the COPY_TICKS enumeration.
        :return: Returns ticks as the numpy array with the named time, bid, ask, last and flags columns.
        The 'flags' value can be a combination of flags from the TICK_FLAG enumeration. Return None in case of
        an error. The info on the error can be obtained using last_error().

        Note:
            See the CopyTicks function for more information.

            When creating the 'datetime' object, Python uses the local time zone, while MetaTrader 5 stores tick
            and bar open time in UTC time zone (without the shift). Therefore, 'datetime' should be created in
            UTC time for executing functions that use time. Data received from the MetaTrader 5 terminal has UTC time.
        """
    symbol = _h.any_symbol(symbol)
    try:
        return mt5_copy_ticks_range(symbol, datetime_from, datetime_to, flags)
    except SystemError:
        return None


mt5_history_deals_get = _mt5.history_deals_get


@_context_manager_modified
def history_deals_get(datetime_from: datetime = None,
                      datetime_to: datetime = None,
                      *,
                      group: str = None,
                      ticket: int = None,
                      position: int = None,
                      function: Callable = None,
                      **kwargs
                      ) -> Tuple[TradeDeal]:
    """Get deals from trading history within the specified interval with the ability to filter by ticket or position.

    :param datetime_from: Date the orders are requested from. Set by the 'datetime' object or as a number of seconds
    elapsed since 1970.01.01.
    :param datetime_to: Date, up to which the orders are requested. Set by the 'datetime' object or as a number of
    seconds elapsed since 1970.01.01.
    :param group: The filter for arranging a group of necessary symbols. Optional named parameter. If the group is
    specified, the function returns only deals meeting a specified criteria for a symbol name.
    :param ticket: Ticket of an order (stored in DEAL_ORDER) all deals should be received for. Optional parameter.
    :param position: Ticket of a position (stored in DEAL_POSITION_ID) all deals should be received for. Optional
    parameter.
    :param function: A function that accepts a TradeDeal object and
    returns True if that object is to be used else False
    :param kwargs:
    :return: a tuple of TradeDeal objects
    """
    args = locals().copy()
    return _h.get_history_type_stuff(mt5_history_deals_get, args)


mt5_history_deals_total = _mt5.history_deals_total


@_context_manager_modified
def history_deals_total(datetime_from: datetime, datetime_to: datetime, **kwargs) -> int:
    """Get the number of ``deals`` in trading history within the specified interval.

    :param datetime_from: Date the orders are requested from. Set by the 'datetime' object or as a number of seconds
    elapsed since 1970.01.01. Required unnamed parameter.
    :param datetime_to: Date, up to which the orders are requested. Set by the 'datetime' object or as a number of
    seconds elapsed since 1970.01.01. Required unnamed parameter.
    :param kwargs:
    :return:
    """
    return mt5_history_deals_total(datetime_from, datetime_to)


mt5_history_orders_get = _mt5.history_orders_get


@_context_manager_modified
def history_orders_get(datetime_from: datetime = None,
                       datetime_to: datetime = None,
                       *,
                       group: str = None,
                       ticket: int = None,
                       position: int = None,
                       function: Callable = None,
                       **kwargs
                       ) -> Tuple[TradeOrder]:
    """Get deals from trading history within the specified interval with the ability to filter by ticket or position.

    :param datetime_from: Date the orders are requested from. Set by the 'datetime' object or as a number of
    seconds elapsed since 1970.01.01.
    :param datetime_to: Date, up to which the orders are requested. Set by the 'datetime' object or as a number of
    seconds elapsed since 1970.01.01.
    :param group: The filter for arranging a group of necessary symbols.
    :param ticket: Ticket of an order (stored in DEAL_ORDER) all deals should be received for. Optional parameter.
    :param position: Ticket of a position (stored in DEAL_POSITION_ID) all deals should be received for.
    Optional parameter.
    :param function: A function that accepts a TradeOrder object and
    returns True if that object is to be used else False
    :param kwargs:
    :return: a tuple of TradeOrder objects
    """
    args = locals().copy()
    return _h.get_history_type_stuff(mt5_history_orders_get, args)


mt5_history_orders_total = _mt5.history_orders_total


@_context_manager_modified
def history_orders_total(datetime_from: datetime, datetime_to: datetime, **kwargs) -> int:
    """Get the number of orders in trading history within the specified interval.

    :param datetime_from: Date the orders are requested from. Set by the 'datetime' object or as a number of seconds
    elapsed since 1970.01.01. Required unnamed parameter.
    :param datetime_to: Date, up to which the orders are requested. Set by the 'datetime' object or as a number of
    seconds elapsed since 1970.01.01. Required unnamed parameter.
    :param kwargs:
    :return:
    """
    return mt5_history_orders_total(datetime_from, datetime_to)


mt5_initialize = _mt5.initialize


@_context_manager_modified
def initialize(path: str = None,
               *,
               login: str = None,
               password: str = None,
               server: str = None,
               portable: bool = False,
               timeout: int = None,
               **kwargs
               ) -> bool:
    """Establish a connection with the MetaTrader 5 terminal. Call without parameters. The terminal for connection
    is found automatically.

    :param path:  Path to the metatrader.exe or metatrader64.exe file. Optional unnamed parameter. It is indicated
    first without a parameter name. If the path is not specified, the module attempts to find the executable file on
    its own.
    :param login: Connection timeout in milliseconds. Optional named parameter. If not specified, the value of
    60 000 (60 seconds) is applied. If the connection is not established within the specified time, the call is
    forcibly terminated and the exception is generated.
    :param password: Trading account password. Optional named parameter. If the password is not set, the password
    saved in the terminal database is applied automatically.
    :param server: Trade server name. Optional named parameter. If no server is set, the last used server is applied
    automatically.
    :param portable: Launch terminal in portable mode
    :timeout: Number of milliseconds for timeout
    :return: Returns True in case of successful connection to the MetaTrader 5 terminal, otherwise - False.
    """
    args = locals().copy()
    args = _h.reduce_args(args)
    result = mt5_initialize(**args)
    return result


mt5_last_error = _mt5.last_error


def last_error() -> Tuple[int, str]:
    """last_error() allows obtaining an error code in case of a failed execution of a MetaTrader 5 library function.
    It is similar to GetLastError(). However, it applies its own error codes.

    :return: Return the last error code and description as a tuple.
    """
    return mt5_last_error()


mt5_login = _mt5.login


@_context_manager_modified
def login(login: int, *,
          password: str = None,
          server: str = None,
          timeout: int = None,
          **kwargs,
          ) -> bool:
    """Connect to a trading account using specified parameters.

    :param login: Trading account number. Required unnamed parameter.
    :param password: Trading account password.
    :param server: Trade server name
    :param timeout: Connection timeout in milliseconds.
    :param kwargs:
    :return: True if success.
    """
    args = locals().copy()
    args = _h.reduce_args(args)
    login = args.pop('login')
    return mt5_login(login, **args)


mt5_order_calc_margin = _mt5.order_calc_margin


@_context_manager_modified
def order_calc_margin(order_type: int,
                      symbol,
                      volume: float,
                      price: float,
                      ) -> float:
    """Return margin in the account currency to perform a specified trading operation.

    :param order_type: Order type taking values from the ORDER_TYPE enumeration
    :param symbol: Financial instrument name.
    :param volume: Trading operation volume.
    :param price: Open price.
    :return: Real value if successful, otherwise None. The error info can be obtained using last_error().
    """
    symbol = _h.any_symbol(symbol)
    return mt5_order_calc_margin(order_type, symbol, volume, price)


mt5_order_calc_profit = _mt5.order_calc_profit


@_context_manager_modified
def order_calc_profit(order_type: int,
                      symbol,
                      volume: float,
                      price_open: float,
                      price_close: float,
                      ) -> float:
    """Return margin in the account currency to perform a specified trading operation.

    :param order_type: Order type taking values from the ORDER_TYPE enumeration
    :param symbol: Financial instrument name.
    :param volume: Trading operation volume.
    :param price_open: Open price.
    :param price_close: Close price.
    :return: Real value if successful, otherwise None. The error info can be obtained using last_error().
    """
    symbol = _h.any_symbol(symbol)
    return mt5_order_calc_profit(order_type, symbol, volume, price_open, price_close)


mt5_order_check = _mt5.order_check


@_context_manager_modified
def order_check(request: dict = None,
                *,
                action: int = None, magic: int = None, order: int = None,
                symbol=None, volume: float = None, price: float = None,
                stoplimit: float = None, sl: float = None, tp: float = None,
                deviation: int = None, type: int = None, type_filling: int = None,
                type_time: int = None, expiration: datetime = None,
                comment: str = None, position: int = None, position_by: int = None,
                **kwargs,
                ) -> OrderCheckResult:
    """Check funds sufficiency for performing a required trading operation. Check result are returned as the
    MqlTradeCheckResult structure.

    :param request: Pass the trade request as a preformed dictionary.
    :param action: Trade operation type
    :param magic: Expert Advisor ID (magic number)
    :param order: Order ticket
    :param symbol: Trade symbol
    :param volume: Requested volume for a deal in lots
    :param price: Price
    :param stoplimit: Stop-limit level
    :param sl: Stop loss level
    :param tp: Take profit level
    :param deviation: Maximum possible deviation from the requested price
    :param type: Order type
    :param type_filling: Order execution type
    :param type_time: Order expiration type
    :param expiration: Order expiration time (for the orders of ORDER_TIME_SPECIFIED type)
    :param comment: Order comment
    :param position: Position ticket
    :param position_by: The ticket of an opposite position
    :param kwargs:
    :return: OrderSendResult namedtuple
    """
    symbol = _h.any_symbol(symbol)
    args = locals().copy()
    return _h.do_trade_action(mt5_order_check, args)


mt5_order_send = _mt5.order_send


@_context_manager_modified
def order_send(request: dict = None,
               *,
               action: int = None, magic: int = None, order: int = None,
               symbol=None, volume: float = None, price: float = None,
               stoplimit: float = None, sl: float = None, tp: float = None,
               deviation: int = None, type: int = None, type_filling: int = None,
               type_time: int = None, expiration: datetime = None,
               comment: str = None, position: int = None, position_by: int = None,
               **kwargs,
               ) -> OrderSendResult:
    """Interaction between the client terminal and a trade server for executing the order placing operation
    is performed by using trade requests. The trade request is represented by the special predefined structure
    of MqlTradeRequest type, which contain all the fields necessary to perform trade deals. The request processing
    result is represented by the structure of MqlTradeResult type.

    :param request: Pass the order request as a dictionary.
    :param action: Trade operation type
    :param magic: Expert Advisor ID (magic number)
    :param order: Order ticket
    :param symbol: Trade symbol
    :param volume: Requested volume for a deal in lots
    :param price: Price
    :param stoplimit: Stop-limit level
    :param sl: Stop loss level
    :param tp: Take profit level
    :param deviation: Maximum possible deviation from the requested price
    :param type: Order type
    :param type_filling: Order execution type
    :param type_time: Order expiration type
    :param expiration: Order expiration time (for the orders of ORDER_TIME_SPECIFIED type)
    :param comment: Order comment
    :param position: Position ticket
    :param position_by: The ticket of an opposite position
    :param kwargs:
    :return: OrderSendResult namedtuple
    """
    args = locals().copy()
    return _h.do_trade_action(mt5_order_send, args)


mt5_orders_get = _mt5.orders_get


@_context_manager_modified
def orders_get(symbol=None,
               *,
               group: str = None,
               ticket: int = None,
               function: Callable = None,
               **kwargs
               ) -> Tuple[TradeOrder]:
    """Get active orders with the ability to filter by symbol or ticket.

    :param symbol: Symbol name. Optional named parameter. If a symbol is specified, the ticket parameter is ignored.
    :param group: The filter for arranging a group of necessary symbols.
    :param ticket: Order ticket (ORDER_TICKET). Optional named parameter.
    :param function: A function that takes a TradeOrder object as its only arg
    and returns truth condition for filtering
    :return: tuple of TradeOrder objects

    Note:
        The function allows receiving all active orders within one call similar to the OrdersTotal and
        OrderSelect tandem.
        The group parameter allows sorting out orders by symbols. '*' can be used at the beginning and the
        end of a string.
        The group parameter may contain several comma separated conditions. A condition can be set as a mask
        using '*'. The logical negation symbol '!' can be used for an exclusion. All conditions are applied
        sequentially, which means conditions of including to a group should be specified first followed by
        an exclusion condition. For example, group="*, !EUR" means that orders for all symbols should be
        selected first and the ones containing "EUR" in symbol names should be excluded afterwards.
    """
    symbol = _h.any_symbol(symbol)
    orders = _h.get_ticket_type_stuff(mt5_orders_get,
                                      symbol=symbol,
                                      group=group,
                                      ticket=ticket,
                                      function=function)
    return orders


mt5_orders_total = _mt5.orders_total


@_context_manager_modified
def orders_total() -> int:
    """Get the number of active orders.

    :return: Integer value.
    """
    return mt5_orders_total()


def period_seconds(timeframe):
    """Get the number of seconds for the respective timeframe

    :param timeframe:
    :return:
    """
    return _const.PERIOD_SECONDS.get(int(timeframe))


mt5_positions_get = _mt5.positions_get


@_context_manager_modified
def positions_get(symbol=None,
                  *,
                  group: str = None,
                  ticket: int = None,
                  function: Callable = None,
                  **kwargs
                  ) -> Tuple[TradePosition]:
    """Get open positions with the ability to filter by symbol or ticket. There are three call options.

    :param symbol:
    :param group:
    :param ticket:
    :param function:
    :return:
    """
    symbol = _h.any_symbol(symbol)
    positions = _h.get_ticket_type_stuff(mt5_positions_get,
                                         symbol=symbol,
                                         group=group,
                                         ticket=ticket,
                                         function=function
                                         )
    return positions


mt5_positions_total = _mt5.positions_total


@_context_manager_modified
def positions_total() -> int:
    """Get the number of open positions.

    :return: Integer value.
    """
    return mt5_positions_total()


mt5_shutdown = _mt5.shutdown


@_context_manager_modified
def shutdown() -> None:
    """Close the previously established connection to the MetaTrader 5 terminal.

    :return: None
    """
    return mt5_shutdown()


mt5_symbol_info = _mt5.symbol_info


@_context_manager_modified
def symbol_info(symbol) -> SymbolInfo:
    """Get data on the specified financial instrument.

    :param symbol:
    :return: Return info in the form of a named tuple structure (namedtuple). Return None in case of an error.
    The info on the error can be obtained using last_error().
    """
    symbol = _h.any_symbol(symbol)
    return mt5_symbol_info(symbol)


mt5_symbol_info_tick = _mt5.symbol_info_tick


@_context_manager_modified
def symbol_info_tick(symbol) -> Tick:
    """Get the last tick for the specified financial instrument.

    :param symbol:
    :return:
    """
    try:
        return _mt5.symbol_info_tick(symbol)
    except Exception:
        symbol = _h.any_symbol(symbol)
        return mt5_symbol_info_tick(symbol)


# direct access to API function without any added overhead


mt5_symbol_select = _mt5.symbol_select


@_context_manager_modified
def symbol_select(symbol, enable: bool = True) -> bool:
    """Select a symbol in the MarketWatch window or remove a symbol from the window.

    :param symbol:
    :param enable:
    :return: True if successful, otherwise – False.
    """
    symbol = _h.any_symbol(symbol)
    return mt5_symbol_select(symbol, enable)


mt5_symbols_get = _mt5.symbols_get


@_context_manager_modified
def symbols_get(*,
                group: str = None,
                regex: str = None,
                function: Callable = None,
                **kwargs
                ) -> Union[Tuple[SymbolInfo], None]:
    """Get all financial instruments from the MetaTrader 5 terminal.
        The group parameter allows sorting out symbols by name. '*' can be used at the beginning and the
        end of a string.
        The group parameter can be used as a named or an unnamed one. Both options work the same way.
        The named option (group="GROUP") makes the code easier to read.
        The group parameter may contain several comma separated conditions. A condition can be set as a
        mask using '*'. The logical negation symbol '!' can be used for an exclusion. All conditions are applied
        sequentially, which means conditions of including to a group should be specified first followed by an
        exclusion condition. For example, group="*, !EUR" means that all symbols should be selected first and
        the ones containing "EUR" in their names should be excluded afterwards.
        Unlike symbol_info(), the symbols_get() function returns data on all requested symbols within a single call.

    :param group: The filter for arranging a group of necessary symbols. Optional parameter. If the group is
    specified, the function returns only symbols meeting a specified criteria.
    :param regex: Regex pattern for symbol filtering.
    :param function: A function that takes a SymbolInfo object as its only arg and returns <bool> for filtering
    the collection of SymbolInfo results.
    :param kwargs:
    :return: A tuple of SymbolInfo objects
    """
    symbols = mt5_symbols_get(group=group) if group else mt5_symbols_get()
    if symbols is None:
        if _state.raise_on_errors:
            build = version()
            if build and build[1] < _const.MIN_TERMINAL_BUILD:
                raise MT5Error(_const.ERROR_CODE.TERMINAL_VERSION_OUTDATED,
                               "The terminal build needs to be updated to support this feature.")
            else:
                error_code, des = last_error()
                if error_code == _const.RES_S_OK:
                    raise MT5Error(_const.ERROR_CODE.UNKNOWN_ERROR,
                                   "Unknown Error. Is the terminal connected?")
        else:
            return None
    if regex:
        if isinstance(regex, str):
            regex = re.compile(regex)
        symbols = filter(lambda s: regex.match(s.name), symbols)
    if function:
        symbols = filter(function, symbols)
    return tuple(symbols)


mt5_symbols_total = _mt5.symbols_total


@_context_manager_modified
def symbols_total() -> int:
    """Get the number of all financial instruments in the MetaTrader 5 terminal. The function is similar to
    SymbolsTotal(). However, it returns the number of all symbols including custom ones and the ones disabled
    in MarketWatch.

    :return: <int>
    """
    return mt5_symbols_total()


mt5_terminal_info = _mt5.terminal_info


@_context_manager_modified
def terminal_info() -> TerminalInfo:
    """Get the connected MetaTrader 5 client terminal status and settings. The function returns all data that can be
    obtained using TerminalInfoInteger, TerminalInfoDouble and TerminalInfoDouble in one call.

    :return: Return info in the form of a named tuple structure (namedtuple). Return None in case of an error.
    The info on the error can be obtained using last_error().
    """
    return mt5_terminal_info()


def trade_retcode_description(retcode):
    try:
        return _const.TRADE_RETCODE(int(retcode)).name
    except (ValueError, AttributeError):
        return "Unknown Trade Retcode"


mt5_version = _mt5.version


@_context_manager_modified
def version() -> Tuple[int, int, str]:
    """Return the MetaTrader 5 terminal version.

    :return: Returns the MetaTrader 5 terminal version, build and release date. Return None in case of an error.
    The info on the error can be obtained using last_error().
    """
    return mt5_version()



import copy

from . import const
from .const import MQL_TRADE_REQUEST_PROPS
from .context import _ContextAwareBase
from .core import order_check
from .core import order_send
from .core import symbol_info_tick
from .helpers import any_symbol
from .helpers import reduce_combine
from .types import *


def create_order_request(request: dict = None, *, action: int = None, magic: int = None,
                         order: int = None, symbol: str = None, volume: float = None,
                         price: float = None, stoplimit: float = None, sl: float = None, tp: float = None,
                         deviation: int = None, type: int = None, type_filling: int = None, type_time: int = None,
                         expiration: int = None, comment: str = None, position: int = None, position_by: int = None,
                         **kwargs) -> dict:
    """Construct a trade request dictionary using the predefined keywords.

    :param request: Pass in an existing request dictionary and the keyword args will update existing or add new.
    :return:
    """
    return Order(**locals().copy()).request()


class Order(_ContextAwareBase):
    __slots__ = MQL_TRADE_REQUEST_PROPS.keys()

    @classmethod
    def as_buy(cls, **kwargs):
        return cls(action=const.TRADE_ACTION.DEAL, type=const.ORDER_TYPE.BUY, **kwargs)

    @classmethod
    def as_sell(cls, **kwargs):
        return cls(action=const.TRADE_ACTION.DEAL, type=const.TRADE_ACTION.DEAL, **kwargs)

    @classmethod
    def as_buy_limit(cls, **kwargs):
        return cls(action=const.TRADE_ACTION.PENDING, type=const.ORDER_TYPE_BUY.LIMIT, **kwargs)

    @classmethod
    def as_sell_limit(cls, **kwargs):
        return cls(action=const.TRADE_ACTION.PENDING, type=const.ORDER_TYPE_SELL.LIMIT, **kwargs)

    @classmethod
    def as_buy_stop(cls, **kwargs):
        return cls(action=const.TRADE_ACTION.PENDING, type=const.ORDER_TYPE_BUY.STOP, **kwargs)

    @classmethod
    def as_sell_stop(cls, **kwargs):
        return cls(action=const.TRADE_ACTION.PENDING, type=const.ORDER_TYPE_SELL.STOP, **kwargs)

    @classmethod
    def as_flatten(cls, position: TradePosition, **kwargs):
        res = cls.as_sell() if position.type == const.ORDER_TYPE_BUY else cls.as_buy()
        return res(position=position.ticket, symbol=position.symbol,
                   volume=position.volume, magic=position.magic, **kwargs)

    @classmethod
    def as_reverse(cls, position: TradePosition, **kwargs):
        res = cls.as_flatten(position)
        res.volume *= 2
        return res

    @classmethod
    def as_adjusted_net_position(cls, position: TradePosition, new_net_position: float, **kwargs):
        volume = position.volume
        volume = -volume if position.type else volume
        new_volume = new_net_position - volume
        order_obj = cls.as_sell(**kwargs) if new_volume < 0.0 else cls.as_buy(**kwargs)
        order_obj.volume = abs(new_volume)
        order_obj.symbol = position.symbol
        order_obj.magic = position.magic
        order_obj.sl = position.sl
        order_obj.tp = position.tp
        return order_obj

    @classmethod
    def as_modify_sltp(cls, position: Union[TradePosition, int], sl=None, tp=None, **kwargs):
        order = cls(
            action=const.TRADE_ACTION.SLTP,
            position=getattr(position, 'ticket', position),
            sl=sl,
            tp=tp,
            **kwargs
        )
        if isinstance(position, TradePosition):
            order.symbol = position.symbol
            order.sl = sl or position.sl
            order.tp = tp or position.tp
        return order

    @classmethod
    def as_delete_pending(cls, order: Union[TradeOrder, int]):
        order_ticket = getattr(order, 'ticket', order)
        order_obj = cls(
            action=const.TRADE_ACTION.REMOVE,
            order=order_ticket,
        )
        return order_obj

    def __init__(self,
                 request: dict = None, *, action: int = None, magic: int = None,
                 order: int = None, symbol=None, volume: float = None,
                 price: float = None, stoplimit: float = None, sl: float = None, tp: float = None,
                 deviation: int = None, type: int = None, type_filling: int = None, type_time: int = None,
                 expiration: int = None, comment: str = None, position: int = None, position_by: int = None,
                 **kwargs
                 ):
        super().__init__()
        args = locals().copy()
        del args['self']
        self.action = self.magic = self.order = self.symbol = self.volume = self.price = None
        self.stoplimit = self.sl = self.tp = self.deviation = self.type = self.type_filling = None
        self.type_time = self.expiration = self.comment = self.position = self.position_by = None
        self.__call__(**args)

    def __call__(self,
                 request: dict = None, *, action: int = None, magic: int = None,
                 order: int = None, symbol=None, volume: float = None,
                 price: float = None, stoplimit: float = None, sl: float = None, tp: float = None,
                 deviation: int = None, type: int = None, type_filling: int = None, type_time: int = None,
                 expiration: int = None, comment: str = None, position: int = None, position_by: int = None,
                 **kwargs
                 ):

        args = locals().copy()
        request = args.pop('request', None) or {}
        symbol = args.get('symbol') or request.get('symbol')
        args['symbol'] = any_symbol(symbol)
        self._set_self_kw(reduce_combine(request, args))
        return self

    def __repr__(self):
        name = type(self).__name__
        args = [f'{k}={v}' for k, v in self.request().items()]
        sig = f"({', '.join(args)})"
        return name + sig

    def _set_self_kw(self, kw: dict):
        for k, v in kw.items():
            if v is not None and k in self.__slots__:
                setattr(self, k, v)

    def request(self) -> dict:
        # TODO check this against a control. Make sure it works when passing request as well as when overriding the rew.
        req = {}
        for k in self.__slots__:
            v = getattr(self, k)
            if v is not None:
                req[k] = v
        # refactor for older python versions
        # req = {k: v for k in self.__slots__ if (v := getattr(self, k)) is not None}
        return req

    def check(self) -> OrderCheckResult:
        # TODO test
        return order_check(self.request())

    def send(self) -> OrderSendResult:
        # TODO test
        req = self.request()
        action = req.get('action')
        if action == const.TRADE_ACTION.DEAL:
            price = req.get('price')
            if price is None:
                symbol = req.get('symbol')
                if symbol:
                    t = req.get('type')
                    tick = symbol_info_tick(symbol)
                    req['price'] = tick.ask if t == const.ORDER_TYPE.BUY else tick.bid
        res = order_send(req)
        return res

    def copy(self) -> 'Order':
        return copy.deepcopy(self)



from . import const
from .context import _ContextAwareBase
from .core import copy_rates_from_pos
from .core import symbol_info
from .core import symbol_info_tick
from .core import symbol_select
from .types import SymbolInfo
from .types import Union


class Symbol(_ContextAwareBase):
    def __init__(self, symbol: Union[str, SymbolInfo]):
        super().__init__()
        self.name = symbol

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, symbol):
        try:
            self._name = symbol.name
            self._info = symbol
        except AttributeError:
            self._name = symbol
            self._info = None
        self._refresh()

    @property
    def select(self):
        return self._select

    @select.setter
    def select(self, enable: bool):
        x = symbol_select(self._name, enable)
        self._select = enable if x else not enable

    @property
    def tick(self):
        return self._tick

    @property
    def spread(self):
        return int(round(self.spread_float / self.trade_tick_size))

    @property
    def bid(self):
        return self.tick.bid

    @property
    def ask(self):
        return self.tick.ask

    @property
    def volume(self):
        return self.tick.volume

    def _daily(self, key):
        rate = copy_rates_from_pos(self.name, const.TIMEFRAME.D1, 0, 1)
        try:
            return rate[0][key]
        except:
            return 0

    @property
    def day_real_volume(self):
        return self._daily('real_volume')

    @property
    def day_volume(self):
        return self._daily('volume')

    @property
    def volume_real(self):
        return self.tick.volume_real

    def normalize_price(self, price: float):
        ts = self.trade_tick_size
        if ts != 0.0:
            return round(round(price / ts) * ts, self.digits)

    def refresh_rates(self):
        self._tick = symbol_info_tick(self.name)
        return self

    def _refresh(self):
        info = self._info or symbol_info(self._name)
        self.refresh_rates()
        self._select = info.select
        # self.spread = info.spread
        # self.volume_real = info.volume_real
        self.custom = info.custom
        self.chart_mode = info.chart_mode
        self.visible = info.visible
        self.session_deals = info.session_deals
        self.session_buy_orders = info.session_buy_orders
        self.session_sell_orders = info.session_sell_orders
        # self.volume = info.volume
        self.volumehigh = info.volumehigh
        self.volumelow = info.volumelow
        self.digits = info.digits
        self.spread_float = info.spread_float
        self.ticks_bookdepth = info.ticks_bookdepth
        self.trade_calc_mode = info.trade_calc_mode
        self.trade_mode = info.trade_mode
        self.start_time = info.start_time
        self.expiration_time = info.expiration_time
        self.trade_stops_level = info.trade_stops_level
        self.trade_freeze_level = info.trade_freeze_level
        self.trade_exemode = info.trade_exemode
        self.swap_mode = info.swap_mode
        self.swap_rollover3days = info.swap_rollover3days
        self.margin_hedged_use_leg = info.margin_hedged_use_leg
        self.expiration_mode = info.expiration_mode
        self.filling_mode = info.filling_mode
        self.order_mode = info.order_mode
        self.order_gtc_mode = info.order_gtc_mode
        self.option_mode = info.option_mode
        self.option_right = info.option_right
        self.bidhigh = info.bidhigh
        self.bidlow = info.bidlow
        self.askhigh = info.askhigh
        self.asklow = info.asklow
        self.lasthigh = info.lasthigh
        self.lastlow = info.lastlow
        self.volumehigh_real = info.volumehigh_real
        self.volumelow_real = info.volumelow_real
        self.option_strike = info.option_strike
        self.point = info.point
        self.trade_tick_value = info.trade_tick_value
        self.trade_tick_value_profit = info.trade_tick_value_profit
        self.trade_tick_value_loss = info.trade_tick_value_loss
        self.trade_tick_size = info.trade_tick_size
        self.trade_contract_size = info.trade_contract_size
        self.trade_accrued_interest = info.trade_accrued_interest
        self.trade_face_value = info.trade_face_value
        self.trade_liquidity_rate = info.trade_liquidity_rate
        self.volume_min = info.volume_min
        self.volume_max = info.volume_max
        self.volume_step = info.volume_step
        self.volume_limit = info.volume_limit
        self.swap_long = info.swap_long
        self.swap_short = info.swap_short
        self.margin_initial = info.margin_initial
        self.margin_maintenance = info.margin_maintenance
        self.session_volume = info.session_volume
        self.session_turnover = info.session_turnover
        self.session_interest = info.session_interest
        self.session_buy_orders_volume = info.session_buy_orders_volume
        self.session_sell_orders_volume = info.session_sell_orders_volume
        self.session_open = info.session_open
        self.session_close = info.session_close
        self.session_aw = info.session_aw
        self.session_price_settlement = info.session_price_settlement
        self.session_price_limit_min = info.session_price_limit_min
        self.session_price_limit_max = info.session_price_limit_max
        self.margin_hedged = info.margin_hedged
        self.price_change = info.price_change
        self.price_volatility = info.price_volatility
        self.price_theoretical = info.price_theoretical
        self.price_greeks_delta = info.price_greeks_delta
        self.price_greeks_theta = info.price_greeks_theta
        self.price_greeks_gamma = info.price_greeks_gamma
        self.price_greeks_vega = info.price_greeks_vega
        self.price_greeks_rho = info.price_greeks_rho
        self.price_greeks_omega = info.price_greeks_omega
        self.price_sensitivity = info.price_sensitivity
        self.basis = info.basis
        self.category = info.category
        self.currency_base = info.currency_base
        self.currency_profit = info.currency_profit
        self.currency_margin = info.currency_margin
        self.bank = info.bank
        self.description = info.description
        self.exchange = info.exchange
        self.formula = info.formula
        self.isin = info.isin
        self.page = info.page
        self.path = info.path
        return self



import enum
import time
from datetime import datetime
from datetime import timedelta

from . import period_seconds
from .const import COPY_TICKS
from .const import TIMEFRAME
from .core import copy_rates_from_pos
from .core import copy_ticks_range
from .core import symbol_info
from .types import *


class EVENT(enum.IntFlag):
    TICK_LAST_CHANGE = enum.auto()
    NEW_BAR = enum.auto()


def iter_event(symbol: Union[str, SymbolInfo],
               timeframe: TIMEFRAME,
               event_flags: Union[EVENT, int],
               sleep: float = 0.001):
    if event_flags == 0:
        return
    symbol = getattr(symbol, 'name', symbol_info(symbol))
    last_tick = None
    last_bar = copy_rates_from_pos(symbol.name, timeframe, 0, 1)[0]
    psec = timedelta(seconds=period_seconds(timeframe))
    next_bar_time = datetime.fromtimestamp(last_bar['time']) + psec
    # save as bool to prevent checking condition on each loop
    do_tick_event = EVENT.TICK_LAST_CHANGE in event_flags
    do_new_bar_event = EVENT.NEW_BAR in event_flags

    while True:
        if do_tick_event:
            now = datetime.now()
            last_tick_time = now - timedelta(seconds=1) if last_tick is None else datetime.fromtimestamp(last_tick.time)
            if last_tick_time > now:
                now, last_tick_time = last_tick_time, now
            ticks = copy_ticks_range(symbol.name, last_tick_time, now, COPY_TICKS.ALL)
            if ticks is not None:
                for tick in ticks:
                    tick = CopyTick(*tick)
                    if last_tick is None or tick.time_msc > last_tick.time_msc:
                        yield EVENT.TICK_LAST_CHANGE, tick
                        last_tick = tick
        if do_new_bar_event:
            if datetime.now() >= next_bar_time:
                bar = copy_rates_from_pos(symbol.name, timeframe, 0, 1)[0]
                if (bar_time := bar['time']) != last_bar['time']:
                    yield EVENT.NEW_BAR, CopyRate(*bar)
                    last_bar = bar
                    next_bar_time = datetime.fromtimestamp(bar_time) + psec
        time.sleep(sleep)



from . import const
from .context import _ContextAwareBase
from .core import positions_get
from .order import Order
from .symbol import Symbol
from .types import *


class Trade(_ContextAwareBase):

    def __init__(self, symbol: Union[str, SymbolInfo], magic=0):
        super().__init__()
        self._symbol = None
        self.symbol = symbol
        self.magic = magic
        # self._order = Order(symbol=self.symbol, magic=self.magic)
        self._position = None

    @property
    def symbol(self) -> Union[Symbol, None]:
        return self._symbol

    @symbol.setter
    def symbol(self, new_symbol):
        if isinstance(new_symbol, (str, Symbol, SymbolInfo)):
            s = Symbol(new_symbol)
        else:
            raise TypeError('Wrong assignment type. Must be str, Symbol, or SymbolInfo')
        self._symbol = s

    @property
    def position(self) -> TradePosition:
        if self._position is None:
            self.refresh()
        return self._position

    def refresh(self):
        p = positions_get(symbol=self.symbol.name, function=lambda p: p.magic == self.magic)
        self._position = p[0] if p else None
        return self

    def _do_market(self, order_constructor, volume: float, comment: str = None, **kwargs) -> OrderSendResult:
        price = self.symbol.refresh_rates().tick.ask
        order = order_constructor(
            symbol=self.symbol,
            magic=self.magic,
            price=price,
            volume=volume,
            comment=comment,
            **kwargs
        )
        result = order.send()
        return result

    def buy(self, volume: float, comment: str = None, **kwargs) -> OrderSendResult:
        return self._do_market(Order.as_buy, volume, comment, **kwargs)

    def sell(self, volume: float, comment: str = None, **kwargs) -> OrderSendResult:
        return self._do_market(Order.as_sell, volume, comment, **kwargs)

    def modify_sltp_by_price(self, sl: int = None, tp: int = None, **kwargs) -> OrderSendResult:
        """Modify an existing position by exact prices.

        :param sl: stop loss price
        :param tp: take profit price
        :param kwargs: kwargs are propagated to the Order class used for the order
        :return: Result of order_send
        """
        p = kwargs.get('position') or self.refresh().position
        if sl:
            sl = self.symbol.normalize_price(sl)
        if tp:
            tp = self.symbol.normalize_price(tp)
        order = Order.as_modify_sltp(position=p, sl=sl, tp=tp)
        result = order.send()
        return result

    def modify_sltp_by_ticks(self,
                             sl: int = None,
                             tp: int = None,
                             price_basis: Union[str, float] = 'open',
                             **kwargs
                             ) -> OrderSendResult:
        """Modify an existing position by number of ticks.

        :param sl: number of ticks away from the defined price for stoploss
        :param tp: number of ticks away from the defined price for take profit
        :param price_basis: the price basis for calculation of levels. if a str == 'open' is passed then the stops
        will be calculated using the position cost-basis. Otherwise, 'current' will use the current price.
        :param kwargs: kwargs are propagated to the Order class used for the order
        :return: Result of order_send
        """
        position = self.refresh().position
        tick_size = self.symbol.trade_tick_size
        if isinstance(price_basis, float):
            bid = ask = price_basis
        elif 'current' in price_basis.lower():
            tick = self.symbol.refresh_rates().tick
            bid, ask = tick.bid, tick.ask
        else:
            bid = ask = position.price_open
        if position.type == const.POSITION_TYPE.BUY:
            price = bid
            stop = -abs(sl) if sl else None
            take = abs(tp) if tp else None
        else:
            price = ask
            take = -abs(tp) if tp else None
            stop = abs(sl) if sl else None
        if stop:
            stop = price + stop * tick_size
        if take:
            take = price + take * tick_size
        result = self.modify_sltp_by_price(sl=stop, tp=take, position=position, **kwargs)
        return result



import MetaTrader5 as _mt5
from collections import namedtuple

from typing import Callable
from typing import Iterable
from typing import Tuple
from typing import Union
from typing import Any
from typing import Optional

# custom namedtuples
CopyRate = namedtuple("CopyRate", "time, open, high, low, close, tick_volume, spread, real_volume")
CopyTick = namedtuple("CopyTick", "time, bid, ask, last, volume, time_msc, flags, volume_real")
# MT5 namedtuple objects for typing
Tick = _mt5.Tick
AccountInfo = _mt5.AccountInfo
SymbolInfo = _mt5.SymbolInfo
TerminalInfo = _mt5.TerminalInfo
OrderCheckResult = _mt5.OrderCheckResult
OrderSendResult = _mt5.OrderSendResult
TradeOrder = _mt5.TradeOrder
TradeDeal = _mt5.TradeDeal
TradeRequest = _mt5.TradeRequest
TradePosition = _mt5.TradePosition



import MetaTrader5 as _mt5

from . import types
from .const import *
from .context import connected
from .context import handle_exit
from .core import *
from .helpers import as_dict_all
from .oem import *

__version__ = _mt5.__version__
__author__ = _mt5.__author__






import MetaTrader5 as _mt5

from .types import *


# internal order send
def _RawOrder(order_type: int,
              symbol: str,
              volume: float,
              price: float,
              comment: str = None,
              ticket: int = None
              ) -> OrderSendResult:
    return _mt5._RawOrder(order_type, symbol, volume, price, comment, ticket)


# Close all specific orders
def Close(symbol: str, *, comment: str = None, ticket: int = None) -> Union[bool, str]:
    return _mt5.Close(symbol, comment=comment, ticket=ticket)


# Buy order
def Buy(symbol: str, volume: float, price: float = None, *,
        comment: str = None, ticket: int = None) -> OrderSendResult:
    return _mt5.Buy(symbol, volume, price, comment=comment, ticket=ticket)


# Sell order
def Sell(symbol: str, volume: float, price: float = None, *,
         comment: str = None, ticket: int = None) -> OrderSendResult:
    return _mt5.Sell(symbol, volume, price, comment=comment, ticket=ticket)
